"use strict";
var __decorate = (this && this.__decorate) || function (decorators, target, key, desc) {
    var c = arguments.length, r = c < 3 ? target : desc === null ? desc = Object.getOwnPropertyDescriptor(target, key) : desc, d;
    if (typeof Reflect === "object" && typeof Reflect.decorate === "function") r = Reflect.decorate(decorators, target, key, desc);
    else for (var i = decorators.length - 1; i >= 0; i--) if (d = decorators[i]) r = (c < 3 ? d(r) : c > 3 ? d(target, key, r) : d(target, key)) || r;
    return c > 3 && r && Object.defineProperty(target, key, r), r;
};
var __metadata = (this && this.__metadata) || function (k, v) {
    if (typeof Reflect === "object" && typeof Reflect.metadata === "function") return Reflect.metadata(k, v);
};
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : new P(function (resolve) { resolve(result.value); }).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
Object.defineProperty(exports, "__esModule", { value: true });
const decorators_1 = require("../decorators");
const Logger_1 = require("../shared/Logger");
const Server_1 = require("../Server");
const errors = require("../errors");
const Token_1 = require("../shared/Token");
const util_1 = require("../shared/util");
const realms_1 = require("../realms");
const httpProxy = require("http-proxy");
let SyncProxyService = class SyncProxyService {
    constructor() {
        this.logger = new Logger_1.Mute();
        this.backends = {};
        this.nextSocketID = 0;
    }
    start(server) {
        return __awaiter(this, void 0, void 0, function* () {
            this.server = server;
        });
    }
    stop() {
        return __awaiter(this, void 0, void 0, function* () {
            for (const label in this.backends) {
                const backend = this.backends[label];
                if (backend) {
                    this.closeBackend(backend);
                    delete this.backends[label];
                }
            }
        });
    }
    setLogger(logger) {
        this.logger = logger;
    }
    syncMasterChanged(event, handle) {
        const labels = handle.tags.filter((t) => t !== 'master');
        if (labels.length !== 1) {
            this.logger.error("master sync service should have exactly one label besides 'master'");
            return;
        }
        const label = labels[0];
        const backend = this.backends[label];
        if (!backend) {
            this.logger.error(`Watch callback for backend that does not exist: ${label}`);
            return;
        }
        if (event === "unavailable") {
            this.closeBackend(backend);
        }
        else if (event === "available") {
            const oldHandle = backend.handle;
            if (oldHandle) {
                if (oldHandle.address === handle.address && oldHandle.port === handle.port) {
                    this.logger.debug("ignore repeated 'available' message from service watch");
                    return;
                }
            }
            this.closeBackend(backend);
            backend.handle = handle;
            const target = `http://${handle.address}:${handle.port}`;
            this.logger.debug(`new target for ${label} is ${target}`);
            backend.sockets = {};
            backend.proxy = httpProxy.createProxy({ ws: true, target });
            backend.proxy.on('error', (err) => {
                this.logger.warn("sync proxy error", err);
            });
            backend.proxy.on('close', () => {
                this.logger.debug("sync proxy closed");
            });
        }
    }
    closeBackend(backend) {
        if (backend.proxy) {
            backend.proxy.close();
            delete backend.proxy;
        }
        if (backend.sockets) {
            for (let id in backend.sockets) {
                const socket = backend.sockets[id];
                socket.destroy();
                this.logger.debug(`destroyed connection from ${socket.remoteAddress}:${socket.remotePort}`);
            }
            backend.sockets = {};
        }
    }
    parseTokenAndGetLabel(req, socket, head) {
        return __awaiter(this, void 0, void 0, function* () {
            const authorization = req.headers['authorization'];
            if (!authorization) {
                throw new errors.JSONError({
                    title: "Authorization header was not provided",
                    status: 400,
                });
            }
            const matches = authorization.match(/^Realm\-Access\-Token version=(\d+) token="(.*)"$/);
            if (!matches) {
                throw new errors.JSONError({
                    title: "Authorization header is not in a valid format",
                    status: 400,
                });
            }
            const tokenVersion = matches[1];
            const tokenData = matches[2];
            let token;
            try {
                token = Token_1.Token.parse(tokenData, this.server.publicKey);
            }
            catch (err) {
                throw new errors.realm.InvalidCredentials();
            }
            const pathParam = util_1.validateRealmPath(req.params.path, token);
            if (!(token instanceof Token_1.AccessToken)) {
                throw new errors.realm.InvalidParameters(`The provided token (${token.data()}) is not an instanceof AccessToken.`);
            }
            const accessToken = token;
            if (accessToken.path !== pathParam && !accessToken.isAdminToken()) {
                throw new errors.realm.InvalidParameters(`The path in the provided token ${accessToken.path} is not the same as that in the URL (${pathParam}).`);
            }
            let label = accessToken.syncLabel;
            if (!label || label === "") {
                label = yield this.findSyncLabelForPath(pathParam, tokenData);
                if (!label) {
                    throw new errors.JSONError({
                        title: "AccessToken is not valid: sync label was not found in token",
                        status: 400,
                    });
                }
            }
            return label;
        });
    }
    websocketHandler(req, socket, head) {
        return __awaiter(this, void 0, void 0, function* () {
            const socketID = this.nextSocketID++;
            this.logger.debug(`[${socketID}] new connection from ${socket.remoteAddress}:${socket.remotePort}`);
            const label = yield this.parseTokenAndGetLabel(req, socket, head);
            const backend = yield this.getBackend(label);
            if (!backend.proxy) {
                throw new errors.JSONError({
                    title: "Backend service is unavailable",
                    status: 503,
                });
            }
            backend.sockets[socketID] = socket;
            socket.on('close', () => {
                this.logger.debug(`[${socketID}] closed connection from ${socket.remoteAddress}:${socket.remotePort}`);
                if (backend.sockets) {
                    delete backend.sockets[socketID];
                }
            });
            backend.proxy.ws(req, socket, head);
        });
    }
    findSyncLabelForPath(path, tokenData) {
        return __awaiter(this, void 0, void 0, function* () {
            const adminRealm = yield this.server.getSyncedRealm(realms_1.AdminRealm).open();
            const realmFile = adminRealm.objectForPrimaryKey("RealmFile", path);
            if (!realmFile) {
                throw new errors.realm.RealmNotFound();
            }
            return realmFile.syncLabel;
        });
    }
    getBackend(syncLabel) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.backends[syncLabel]) {
                const watch = this.server.discovery.watchService("sync", ["master", syncLabel]);
                watch.on('available', (handle) => this.syncMasterChanged("available", handle));
                watch.on('unavailable', (handle) => this.syncMasterChanged("unavailable", handle));
                this.backends[syncLabel] = { watch };
                const handle = yield this.server.discovery.find("sync", ["master", syncLabel]);
                if (handle) {
                    this.syncMasterChanged("available", handle);
                }
            }
            return this.backends[syncLabel];
        });
    }
};
__decorate([
    decorators_1.Start(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Server_1.Server]),
    __metadata("design:returntype", Promise)
], SyncProxyService.prototype, "start", null);
__decorate([
    decorators_1.Stop(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", []),
    __metadata("design:returntype", Promise)
], SyncProxyService.prototype, "stop", null);
__decorate([
    decorators_1.Unmute(),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Logger_1.Logger]),
    __metadata("design:returntype", void 0)
], SyncProxyService.prototype, "setLogger", null);
__decorate([
    decorators_1.Upgrade('/:path'),
    __metadata("design:type", Function),
    __metadata("design:paramtypes", [Object, Object, Object]),
    __metadata("design:returntype", Promise)
], SyncProxyService.prototype, "websocketHandler", null);
SyncProxyService = __decorate([
    decorators_1.BaseRoute('/realm-sync')
], SyncProxyService);
exports.SyncProxyService = SyncProxyService;
//# sourceMappingURL=SyncProxyService.js.map